#
# mobo.rb - 󤬲ȷ Library
#
# Copyright (C) 2003, Kazuhiko <kazuhiko@fdiary.net>
#

MOBO_VERSION = '0.3'

require 'pstore'
require 'cgi'
require 'date'
begin
	require 'erb_fast'
	ERbLight = ERB
rescue LoadError
	begin
		require 'erb'
		ERbLight = ERB
	rescue LoadError
		require 'erb/erbl'
	end
end

#
# Integer
#
class Integer
	# adding comma each three digits.
	def to_disp_str
		self.to_s.reverse.gsub( /\d\d\d/, '\0,' ).reverse.sub( /^([-]{0,1}),/, '\1' )
	end
end

#
# CGI
#
class CGI
	def valid?( param, idx = 0 )
		begin
			self.params[param] and self.params[param][idx] and self.params[param][idx].length > 0
		rescue NameError # for Tempfile class of ruby 1.6
			self.params[param][idx].stat.size > 0
		end
	end

	def mobile_agent?
		self.user_agent =~ %r[(DoCoMo|J-PHONE|UP\.Browser|DDIPOCKET|ASTEL|PDXGW|Palmscape|Xiino|sharp pda browser|Windows CE|L-mode)]i
	end
end

#
# Safe Module
#
require 'thread'
module Safe
	def safe( level = 4 )
		result = nil
		if $SAFE < level then
			Thread.start {
				$SAFE = level
				result = yield
			}.join
		else
			result = yield
		end
		result
	end
	module_function :safe
end

#
# to_icalc Method
#
class String
	def to_icalc
		if self =~ /[^0-9.+\-*\/(), ]/ then
			return 0
		else
			begin
				return eval( self.gsub( /[, ]/, '' )).to_i
			rescue ZeroDivisionError, SyntaxError
				return 0
			end
		end
	end
end

class NilClass
	def to_icalc
		return 0
	end
end

#
# MoBo Class
#
module MoBo
	PATH = File::dirname( __FILE__ )

	#
	# class Item
	#
	class Item
		attr_reader :date, :genre, :price, :note, :account
		def initialize( date, genre, price, note, account, special = nil)
			raise ArgumentError::new( "Bad date [#{date}]" ) if not date
			raise ArgumentError::new( 'Bad price' ) if not price or price == 0
			@date = date.to_i
			@genre = genre
			@price = price
			@note = note
			@account = account
			@special = special
		end

		def special
			if @special == 'on' then
				'*'
			else
				''
			end
		end

		def <=>( other )
			return @genre <=> other.genre
		end
	end

	#
	# class Daily
	#
	class Daily
		attr_reader :date
		def initialize( date )
			raise ArgumentError::new( 'Bad date' ) if (/^\d{8}$/ !~ date) == nil
			@date = date
			@items = []
		end

		def add( item )
			@items << item
		end

		def each_item( cond = {} )
			@items.sort.each do |i|
				flag = true
				cond['account'] = // unless cond['account']
				cond.each do |key, val|
					if val.class == String
						flag = flag && (val == i.send(key))
					else
						flag = flag && (val =~ i.send(key))
					end
				end
				yield i if flag
			end
		end

		def item_size
			@items.length
		end
	end

	#
	# class Monthly
	#
	class Monthly
		attr_reader :month
		def initialize( month )
			raise ArgumentError::new( 'Bad month' ) if (/^\d{6}$/ !~ month) == nil
			@month = month
			@diaries = {}
		end

		def add( item )
			date = "#{month}#{sprintf('%02d',item.date)}"
			diary = @diaries[date] || Daily.new( date )
			diary.add( item )
			@diaries[date] = diary
		end

		def []( date )
			@diaries[date]
		end

		def []=( date, diary )
			@diaries[date] = diary
		end

		def each_diary( cond = {} )
			@diaries.sort.each do |date,diary|
				yield diary
			end
		end

		def each_item( cond = {} )
			@diaries.sort.each do |date,diary|
				diary.each_item( cond ) do |i|
					yield i
				end
			end
		end

		def each_notmove_item( cond = {} )
			cond['genre'] = /^[ie]:/ unless cond['genre']
			each_item( cond ) do |i|
				yield i
			end
		end

		def each_move_item( cond = {} )
			return if cond['genre']
			cond['genre'] = /^m:/
			each_item( cond ) do |i|
				yield i
			end
		end

		def item_size
			total = 0
			each_item do
				total += 1
			end
			total
		end

		def total( cond = {} )
			price = 0
			each_item(cond) { |i| price += i.price }
			price
		end
	end

	# class ForceRedirect
	#  force redirect to another page
	#
	class ForceRedirect < StandardError
		attr_reader :path
		def initialize( path )
			@path = path
		end
	end

	#
	# class Config
	#
	class Config
		def initialize
			load

			instance_variables.each do |v|
				v.sub!( /@/, '' )
				instance_eval( <<-SRC
					def #{v}
						@#{v}
					end
					def #{v}=(p)
						@#{v} = p
					end
					SRC
				)
			end
		end

		# saving to mobo.conf in @data_path
		def save
			result = ERbLight::new( File::open( "#{PATH}/skel/mobo.rconf" ){|f| f.read }.untaint ).result( binding )
			result.untaint unless @secure
			Safe::safe( @secure ? 4 : 1 ) do
				eval( result )
			end
			File::open( "#{@data_path}mobo.conf", 'w' ) do |o|
				o.print result
			end
		end

		def mobile_agent?
			%r[(DoCoMo|J-PHONE|UP\.Browser|DDIPOCKET|ASTEL|PDXGW|Palmscape|Xiino|sharp pda browser|Windows CE|L-mode)]i =~ ENV['HTTP_USER_AGENT']
		end

		#
		# get/set/delete plugin options
		#
		def []( key )
			@options[key]
		end

		def []=( key, val )
			@options2[key] = @options[key] = val
		end

		def delete( key )
			@options.delete( key )
			@options2.delete( key )
		end
	
		def base_url
			if ENV['HTTPS']
				port = (ENV['SERVER_PORT'] == '443') ? '' : ':' + ENV['SERVER_PORT'].to_s
				"https://#{ ENV['SERVER_NAME'] }#{ port }#{File::dirname(ENV['SCRIPT_NAME'])}/"
			else
				port = (ENV['SERVER_PORT'] == '80') ? '' : ':' + ENV['SERVER_PORT'].to_s
				"http://#{ ENV['SERVER_NAME'] }#{ port }#{File::dirname(ENV['SCRIPT_NAME'])}/"
			end.sub(%r|/+$|, '/')
		end

	private
		# loading mobo.conf in current directory
		def load
			@secure = true unless @secure
			@options = {}
			eval( File::open( "mobo.conf" ){|f| f.read }.untaint )

			# language setup
			@lang = 'ja' unless @lang
			begin
				instance_eval( File::open( "#{MoBo::PATH}/mobo/lang/#{@lang}.rb" ){|f| f.read }.untaint )
			rescue Errno::ENOENT
				@lang = 'ja'
				retry
			end

			@data_path += '/' if /\/$/ !~ @data_path
			@index = './' unless @index
			@update = 'update.rb' unless @update

			@index_page = '' unless @index_page
			@hour_offset = 0 unless @hour_offset

			@html_title = '' unless @html_title
			@header = '' unless @header
			@footer = '' unless @footer

			@theme = 'default' if not @theme and not @css
			@css = '' unless @css

			@genre_income = [] unless @genre_income
			@genre_expense = [] unless @genre_expense
			@genre_move = [] unless @genre_move
			@account_list = [] unless @account_list

			@vaccount_name = [] unless @vaccount_name
			@vaccount_list = [] unless @vaccount_list
			@options = {} unless @options.class == Hash
			if @options2 then
				@options.update( @options2 )
			else
				@options2 = {}.taint
			end
			@options.taint
		end

		# loading mobo.conf in @data_path.
		def load_cgi_conf
			raise MoBoError, 'No @data_path variable.' unless @data_path
	
			@data_path += '/' if /\/$/ !~ @data_path
			raise MoBoError, 'Do not set @data_path as same as MoBo system directory.' if @data_path == "#{PATH}/"
	
			variables = [
				:index_page, :hour_offset,
				:html_title, :header, :footer,
				:theme, :css,
				:genre_income, :genre_expense, :genre_move, :account_list,
				:vaccount_name, :vaccount_list,
				:options2,
			]
			begin
				cgi_conf = File::open( "#{@data_path}mobo.conf" ){|f| f.read }
				cgi_conf.untaint unless @secure
				def_vars = ""
				variables.each do |var| def_vars << "#{var} = nil\n" end
				eval( def_vars )
				Safe::safe( @secure ? 4 : 1 ) do
					eval( cgi_conf )
				end
				variables.each do |var| eval "@#{var} = #{var} if #{var} != nil" end
			rescue IOError, Errno::ENOENT
			end
		end

		def method_missing( *m )
			if m.length == 1 then
				instance_eval( <<-SRC
					def #{m[0]}
						@#{m[0]}
					end
					def #{m[0]}=( p )
						@#{m[0]} = p
					end
					SRC
				)
			end
			nil
		end
	end

	#
	# class Plugin
	#  plugin management class
	#
	class Plugin
		attr_reader :cookies

		def initialize( params )
			@header_procs = []
			@footer_procs = []
			@onload_procs = []
			@update_procs = []
			@body_enter_procs = []
			@body_leave_procs = []
			@edit_procs = []
			@form_procs = []
			@conf_keys = []
			@conf_procs = {}
			@menu_procs = []
			@plugin_procs = {}
			@cookies = []

			params.each_key do |key|
				eval( "@#{key} = params['#{key}']" )
			end

			# for ruby 1.6.x support
			if @conf.secure then
				@cgi.params.each_value do |p|
					p.each {|v| v.taint}
				end
			end

			# loading plugins
			@plugin_files = []
			plugin_path = @conf.plugin_path || "#{PATH}/plugin"
			plugin_file = ''
			begin
				Dir::glob( "#{plugin_path}/*.rb" ).sort.each do |file|
					plugin_file = file
					load_plugin( file )
					@plugin_files << plugin_file
				end
			rescue Exception
				raise PluginError::new( "Plugin error in '#{File::basename( plugin_file )}'.\n#{$!}" )
			end
		end

		def load_plugin( file )
			@resource_loaded = false
			begin
				res_file = File::dirname( file ) + "/#{@conf.lang}/" + File::basename( file )
				open( res_file.untaint ) do |src|
					instance_eval( src.read.untaint )
				end
				@resource_loaded = true
			rescue IOError, Errno::ENOENT
			end
			open( file.untaint ) do |src|
				instance_eval( src.read.untaint )
			end
		end

		def eval_src( src, secure )
			self.taint
			@conf.taint
			@body_enter_procs.taint
			@body_leave_procs.taint
			return Safe::safe( secure ? 4 : 1 ) do
				eval( src )
			end
		end

	private
		def add_header_proc( block = Proc::new )
			@header_procs << block
		end

		def header_proc
			r = []
			@header_procs.each do |proc|
				r << proc.call
			end
			r.join.chomp
		end

		def add_footer_proc( block = Proc::new )
			@footer_procs << block
		end

		def footer_proc
			r = []
			@footer_procs.each do |proc|
				r << proc.call
			end
			r.join.chomp
		end

		def add_onload_proc( block = Proc::new )
			@onload_procs << block
		end

		def onload_proc
			r = []
			@onload_procs.each do |proc|
				r << proc.call
			end
			r.join.strip
		end

		def add_update_proc( block = Proc::new )
			@update_procs << block
		end

		def update_proc
			@update_procs.each do |proc|
				proc.call
			end
			''
		end

		def add_body_enter_proc( block = Proc::new )
			@body_enter_procs << block
		end

		def body_enter_proc( date )
			r = []
			@body_enter_procs.each do |proc|
				r << proc.call( date )
			end
			r.join
		end

		def add_body_leave_proc( block = Proc::new )
			@body_leave_procs << block
		end

		def body_leave_proc( date )
			r = []
			@body_leave_procs.each do |proc|
				r << proc.call( date )
			end
			r.join
		end

		def add_edit_proc( block = Proc::new )
			@edit_procs << block
		end

		def edit_proc( date )
			r = []
			@edit_procs.each do |proc|
				r << proc.call( date )
			end
			r.join
		end

		def add_form_proc( block = Proc::new )
			@form_procs << block
		end

		def form_proc( date )
			r = []
			@form_procs.each do |proc|
				r << proc.call( date )
			end
			r.join
		end

		def add_conf_proc( key, label, block = Proc::new )
			return unless @mode =~ /^(conf|saveconf)$/
			@conf_keys << key unless @conf_keys.index( key )
			@conf_procs[key] = [label, block]
		end

		def each_conf_key
			@conf_keys.each do |key|
				yield key
			end
		end

		def conf_proc( key )
			r = ''
			label, block = @conf_procs[key]
			r = block.call if block
			r
		end

		def conf_label( key )
			label, block = @conf_procs[key]
			label
		end

		def add_menu_proc( block = Proc::new )
			@menu_procs << block
		end

		def menu_proc
			r = []
			@menu_procs.each do |proc|
				r << proc.call
			end
			r.compact
		end

		def add_plugin_proc( key, block = Proc::new )
			return unless @mode == 'plugin'
			@plugin_procs[key] = block
		end

		def plugin_proc( key )
			r = ''
			block = @plugin_procs[key]
			r = block.call if block
			ERbLight.new( r ).result( binding )
		end

		def add_cookie( cookie )
			begin
				@cookies << cookie
			rescue SecurityError
				raise SecurityError, "can't use cookies in plugin when secure mode"
			end
		end

		def apply_plugin( str, remove_tag = false )
			r = str.dup
			if @conf.options['apply_plugin'] and str.index( '<%' ) then
				r = str.untaint if $SAFE < 3
				Safe::safe( @conf.secure ? 4 : 1 ) do
					r = ERbLight.new( r ).result( binding )
				end
			end
			r.gsub!( /<.*?>/, '' ) if remove_tag
			r
		end

		def method_missing( *m )
			super if @debug
			# ignore when no plugin
		end
	end

	#
	# exception classes
	#
	class MoBoError < StandardError; end
	class PermissionError < MoBoError; end
	class PluginError < MoBoError; end

	#
	# class Mobo
	#
	class MoBoBase
		attr_reader :year, :month, :account, :conf, :cookies, :genre
		attr_reader :genre_expense, :genre_income, :genre_move, :genre_list
		attr_reader :account_list, :html_title
		attr_reader :vaccount_name, :vaccount_list

		def initialize(cgi, rhtml, conf)
			@cgi, @rhtml, @conf = cgi, rhtml, conf
			@genre_expense = @conf.genre_expense
			@genre_income = @conf.genre_income
			@genre_move = @conf.genre_move
			@account_list = @conf.account_list
			@vaccount_name = @conf.vaccount_name
			@vaccount_list = @conf.vaccount_list
			@genre_list = []
			@cookies = []
			genre_expense.each do |genre|
				genre_list.push( "e:#{genre}" )
			end
			genre_income.each do |genre|
				genre_list.push( "i:#{genre}" )
			end
			genre_move.each do |genre|
				genre_list.push( "m:#{genre}" )
			end

			@monthly = nil
			@account = @cgi.params['account'][0]
			@genre = @cgi.params['genre'][0]
			@cache = {}
			@date = ''
			now = Time::now
			
			if @cgi.valid?( 'date' ) then
				@year = @cgi.params['date'][0][0,4]
				@month = @cgi.params['date'][0][4,2]
				@month = '01' if @month == ''
			end
			if !Date.exist?(@year.to_i, @month.to_i, 1) then
				@year = now.strftime( '%Y' )
				@month = now.strftime( '%m' )
			end
			@db = PStore::new( "#{@conf.data_path}#{@year}#{@month}.db" )
			@db.transaction do
				begin @monthly = @db['mobo']; rescue; end
				unless @monthly then
					@monthly = Monthly::new( "#{@year}#{@month}" )
					@db.abort
				end
			end
		end

		def monthly( month )
			result = false
			if month == "#{@year}#{@month}" then
				result = @monthly
			elsif FileTest::exist?( "#{@conf.data_path}#{month}.db" ) then
				PStore::new( "#{@conf.data_path}#{month}.db" ).transaction do |db|
					begin result = db['mobo']; rescue; end
				end
			end
			if result then
				result
			else
				Monthly::new( month )
			end
		end

		def month_list
			result = []
			Dir::glob( "#{@conf.data_path}??????.db" ).each do |f|
				result.push( f.gsub( /.*(......)\.db/, '\1' ) )
			end
			result
		end

		def each_monthly
			month_list.each do |m|
				yield monthly( m )
			end
		end

		def genre_total( month, genre, all = nil, account = nil )
			special = all ? // : /^$/
			total( month, 'genre' => genre, 'special' => special, 'account' => account_regexp( account ) )
		end

		def income_total( month, all = nil, account = nil )
			special = all ? // : /^$/
			total( month, 'genre' => /^i:/, 'special' => special, 'account' => account_regexp( account ) )
		end

		def expense_total( month, all = nil, account = nil )
			special = all ? // : /^$/
			total( month, 'genre' => /^e:/, 'special' => special, 'account' => account_regexp( account ) )
		end

		def total( month, cond = {} )
                        cond_str = cond.inspect
                        if !@cache[month] then
                                @c = PStore::new( "#{@conf.data_path}#{month}_cache.db" )
                                @c.transaction do
                                        begin @cache[month] = @c['mobo']; rescue; end
                                        unless @cache[month] then
                                                @cache[month] = Hash::new
                                        end
                                end
                        end
                        if @cache[month][cond_str] then
				#STDERR.puts "cache hit:#{month}"
                                return @cache[month][cond_str]
                        end
			#STDERR.puts "cache miss:#{month}"
                        price = monthly(month).total(cond)
                        @cache[month][cond_str] = price
			@c = PStore::new( "#{@conf.data_path}#{month}_cache.db" )
                        @c.transaction do
                                @c['mobo'] = @cache[month]
                        end
			price
                end

		def grand_total( account = nil )
			price = 0
			month_list.each do |m|
				price += total( m, 'account' => account_regexp( account ) )
			end
			price
		end

		def balance( date, account = nil )
			price = 0
			if /^\d{6}$/ =~ date then
				month_list.each do |m|
					price += total( m, 'account' => account_regexp( account ) ) if m <= date
				end
			elsif /^\d{8}$/ =~ date then
				month = date[0,6]
				day = date[6,2]
				month_list.each do |m|
					price += total( m, 'account' => account_regexp( account ) ) if m < month
				end
				monthly( month ).each_item( 'account' => account_regexp( account ) ) do |i|
					price += i.price if i.date <= day.to_i
				end
			end
			price
		end

		def account_regexp( account )
			return nil unless account
			if i = @vaccount_name.index( account ) then
				return Regexp.new( '^(' + @vaccount_list[i].collect { |v| Regexp.quote( v ) }.join('|') + ')$' )
			else
				return account
			end
		end

		def eval_rhtml( prefix = '' )
			begin
				r = do_eval_rhtml( prefix )
			rescue PluginError, SyntaxError, ArgumentError
				r = ERbLight::new( File::open( "#{PATH}/skel/plugin_error.rhtml" ) {|f| f.read }.untaint ).result( binding )
			rescue Exception
				raise
			end
			return r
		end

	protected
		def do_eval_rhtml( prefix )
			# load plugin files
			load_plugins

			# load and apply rhtmls
			files = ["header.rhtml", @rhtml, "footer.rhtml"]
			rhtml = files.collect {|file|
				path = "#{PATH}/skel/#{prefix}#{file}"
				begin
					File::open( "#{path}.#{@conf.lang}" ) {|f| f.read }
				rescue
					File::open( path ) {|f| f.read }
				end
			}.join
			r = ERbLight::new( rhtml.untaint ).result( binding )
			r = ERbLight::new( r ).src

			# apply plugins
			r = @plugin.eval_src( r.untaint, @conf.secure ) if @plugin
			@cookies += @plugin.cookies
			r
		end

		def mode
			self.class.to_s.sub( /^MoBo::MoBo/, '' ).downcase
		end

		def load_plugins
			calendar
			@plugin = Plugin::new(
				'conf' => @conf,
				'mode' => mode,
				'cgi' => @cgi,
				'years' => @years,
				'date' => "#{@year}#{@month}",
				'year' => @year,
				'month' => @month,
				'monthly' => @monthly,
				'account' => @account
			)
		end

		def save_items
			begin
				File::delete( "#{@conf.data_path}#{@year}#{@month}_cache.db" )
			rescue Errno::ENOENT
			end
			@db.transaction do
				1.upto( 999 ) do |i|
					next if !Date.exist?(@year.to_i, @month.to_i, @cgi.params["date#{i}"][0].to_i)
					price = @cgi.params["price_in#{i}"][0].to_icalc - @cgi.params["price_ex#{i}"][0].to_icalc
					if price != 0 then
						if !@cgi.params["account_in#{i}"][0] then
							item = Item::new( @cgi.params["date#{i}"][0], @conf.to_native( @cgi.params["genre#{i}"][0] ), price, @conf.to_native( @cgi.params["note#{i}"][0] ), @conf.to_native( @cgi.params["account#{i}"][0] ), @cgi.params["special#{i}"][0] )
							@monthly.add( item )
						else
							if @cgi.valid?( "account_in#{i}" ) then
								item = Item::new( @cgi.params["date#{i}"][0], @conf.to_native( @cgi.params["genre#{i}"][0] ), price, @conf.to_native( @cgi.params["note#{i}"][0] ), @conf.to_native( @cgi.params["account_in#{i}"][0] ), @cgi.params["special#{i}"][0] )
								@monthly.add( item )
							end
							if @cgi.valid?( "account_ex#{i}" ) then
								item = Item::new( @cgi.params["date#{i}"][0], @conf.to_native( @cgi.params["genre#{i}"][0] ), -price, @conf.to_native( @cgi.params["note#{i}"][0] ), @conf.to_native( @cgi.params["account_ex#{i}"][0] ), @cgi.params["special#{i}"][0] )
								@monthly.add( item )
							end
						end
					end
				end
				@db['mobo'] = @monthly
			end
		end

		def calendar
			@years = {}
			month_list.sort.each do |m|
				year = m[0, 4]
				if monthly( m ).item_size > 0 then
					@years[year] = [] unless @years[year]
					@years[year] << m
				end
			end
		end
	end

	#
	# class MoBoMonth
	#  show month mode view
	#
	class MoBoMonth < MoBoBase; end

	#
	# class MoBoAdd
	#  add new items
	#
	class MoBoAdd < MoBoBase
		def initialize( cgi, rhtml, conf )
			super
			save_items
		end

	protected
		def do_eval_rhtml( prefix )
			raise ForceRedirect::new( "#{@conf.index}?date=#{@year}#{@month}" )
		end
	end

	#
	# class MoBoEdit
	#  show edit form
	#
	class MoBoEdit < MoBoBase; end

	#
	# class MoBoSaveMonth
	#  save monthly items
	#
	class MoBoSaveMonth < MoBoBase
		def initialize( cgi, rhtml, conf )
			super
			@monthly = Monthly::new( "#{@year}#{@month}" )
			save_items
		end

	protected
		def do_eval_rhtml( prefix )
			raise ForceRedirect::new( "#{@conf.index}?date=#{@year}#{@month}" )
		end
	end

	#
	# class MoBoConf
	#  show configuration form
	#
	class MoBoConf < MoBoBase
		def initialize( cgi, rhtml, conf )
			super
			@key = @cgi.params['conf'][0]
		end
	end

	#
	# class MoBoSaveConf
	#  save configuration
	#
	class MoBoSaveConf < MoBoConf
		def initialize( cgi, rhtml, conf )
			super
		end

		def eval_rhtml( prefix = '' )
			r = super

			begin
				@conf.save
			rescue
				@error = [$!.dup, $@.dup]
			end

			r
		end
	end

	#
	# class MoBoReport
	#  show report page
	#
	class MoBoReport < MoBoBase
		def initialize( cgi, rhtml, conf )
			super
		end

		def all?
			@cgi.valid?( 'all' )
		end
	end

	#
	# class MoBoPlugin
	#  show plugin result
	#
	class MoBoPlugin < MoBoBase
		def initialize( cgi, rhtml, conf )
			super
		end
	end
end
